/*
 $Id: GString.java,v 1.3 2003/11/11 12:17:31 jstrachan Exp $

 Copyright 2003 (C) James Strachan and Bob Mcwhirter. All Rights Reserved.

 Redistribution and use of this software and associated documentation
 ("Software"), with or without modification, are permitted provided
 that the following conditions are met:

 1. Redistributions of source code must retain copyright
    statements and notices.  Redistributions must also contain a
    copy of this document.

 2. Redistributions in binary form must reproduce the
    above copyright notice, this list of conditions and the
    following disclaimer in the documentation and/or other
    materials provided with the distribution.

 3. The name "groovy" must not be used to endorse or promote
    products derived from this Software without prior written
    permission of The Codehaus.  For written permission,
    please contact info@codehaus.org.

 4. Products derived from this Software may not be called "groovy"
    nor may "groovy" appear in their names without prior written
    permission of The Codehaus. "groovy" is a registered
    trademark of The Codehaus.

 5. Due credit should be given to The Codehaus -
    http://groovy.codehaus.org/

 THIS SOFTWARE IS PROVIDED BY THE CODEHAUS AND CONTRIBUTORS
 ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 THE CODEHAUS OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 */
package groovy.lang;


/**
 * Represents a String which contains embedded values such as 
 * "hello there ${user} how are you?"
 * which can be evaluated lazily.
 * Advanced users can iterate over the text and values to perform
 * special processing, such as for performing SQL operations, the
 * values can be substituted for ? and the actual value objects
 * can be bound to a JDBC statement.
 * The lovely name of this class was suggested by Jules Gosnell
 * and was such a good idea, I couldn't resist :)
 * 
 * @author <a href="mailto:james@coredevelopers.net">James Strachan</a>
 * @version $Revision: 1.3 $
 */
public abstract class GString extends GroovyObjectSupport {

    private Object[] values;

    public GString(Object values) {
        this.values = (Object[]) values;
    }
    
    public GString(Object[] values) {
        this.values = values;
    }
    
    // will be static in an instance
    public abstract String[] getStrings();
    
    public Object[] getValues() {
        return values;
    }
    
    public int getValueCount() {
        return values.length;
    }

    public Object getValue(int idx) {
        return values[idx];
    }
    
    public String toString() {
        String[] s = getStrings();
        int numberOfValues = values.length;
        StringBuffer buffer = new StringBuffer();
        for (int i = 0, size = s.length; i < size; i++) {
            buffer.append(s[i]);
            if (i < numberOfValues) {
                buffer.append(values[i]);
            }
        }
        return buffer.toString();
    }
}
